package java.util.concurrent;

import java.util.concurrent.locks.LockSupport;
import sun.misc.Unsafe;

public class FutureTask<V> implements RunnableFuture<V>
{
    private volatile int state;
    private static final int NEW = 0;
    private static final int COMPLETING = 1;
    private static final int NORMAL = 2;
    private static final int EXCEPTIONAL = 3;
    private static final int CANCELLED = 4;
    private static final int INTERRUPTING = 5;
    private static final int INTERRUPTED = 6;
    private Callable<V> callable;
    private Object outcome;
    private volatile Thread runner;
    private volatile FutureTask.WaitNode waiters;
    private static final Unsafe UNSAFE;
    private static final long stateOffset;
    private static final long runnerOffset;
    private static final long waitersOffset;
    
    private V report(int n) throws ExecutionException {
        Object outcome = this.outcome;
        if (n == 2) {
            return outcome;
        }
        if (n >= 4) {
            throw new CancellationException();
        }
        throw new ExecutionException((Throwable)outcome);
    }
    
    public FutureTask(Callable<V> callable) {
        if (callable == null) {
            throw new NullPointerException();
        }
        this.callable = callable;
        this.state = 0;
        return;
    }
    
    public FutureTask(Runnable runnable, V v) {
        this.callable = Executors.callable(runnable, v);
        this.state = 0;
    }
    
    @Override
    public boolean isCancelled() {
        return this.state >= 4;
    }
    
    @Override
    public boolean isDone() {
        return this.state != 0;
    }
    
    @Override
    public boolean cancel(boolean b) {
        if (this.state != 0 || !FutureTask.UNSAFE.compareAndSwapInt(this, FutureTask.stateOffset, 0, b ? 5 : 4)) {
            return false;
        }
        try {
            if (b) {
                try {
                    Thread runner = this.runner;
                    if (runner != null) {
                        runner.interrupt();
                    }
                }
                finally {
                    FutureTask.UNSAFE.putOrderedInt(this, FutureTask.stateOffset, 6);
                }
            }
        }
        finally {
            this.finishCompletion();
        }
        return true;
    }
    
    @Override
    public V get() throws InterruptedException, ExecutionException {
        int n = this.state;
        if (n <= 1) {
            n = this.awaitDone(false, 0L);
        }
        return this.report(n);
    }
    
    @Override
    public V get(long n, TimeUnit timeUnit) throws InterruptedException, ExecutionException, TimeoutException {
        if (timeUnit == null) {
            throw new NullPointerException();
        }
        int n2 = this.state;
        if (n2 <= 1 && (n2 = this.awaitDone(true, timeUnit.toNanos(n))) <= 1) {
            throw new TimeoutException();
        }
        return this.report(n2);
    }
    
    protected void done() {
    }
    
    protected void set(V outcome) {
        if (FutureTask.UNSAFE.compareAndSwapInt(this, FutureTask.stateOffset, 0, 1)) {
            this.outcome = outcome;
            FutureTask.UNSAFE.putOrderedInt(this, FutureTask.stateOffset, 2);
            this.finishCompletion();
        }
    }
    
    protected void setException(Throwable outcome) {
        if (FutureTask.UNSAFE.compareAndSwapInt(this, FutureTask.stateOffset, 0, 1)) {
            this.outcome = outcome;
            FutureTask.UNSAFE.putOrderedInt(this, FutureTask.stateOffset, 3);
            this.finishCompletion();
        }
    }
    
    @Override
    public void run() {
        if (this.state != 0 || !FutureTask.UNSAFE.compareAndSwapObject(this, FutureTask.runnerOffset, null, Thread.currentThread())) {
            return;
        }
        try {
            Callable<Object> callable = this.callable;
            if (callable != null && this.state == 0) {
                Object call;
                int n;
                try {
                    call = callable.call();
                    n = true;
                }
                catch (Throwable exception) {
                    call = null;
                    n = false;
                    this.setException(exception);
                }
                if (!(n == 0)) {
                    this.set(call);
                }
            }
            this.runner = null;
            int state = this.state;
            if (state >= 5) {
                this.handlePossibleCancellationInterrupt(state);
            }
        }
        finally {
            this.runner = null;
            int state2 = this.state;
            if (state2 >= 5) {
                this.handlePossibleCancellationInterrupt(state2);
            }
        }
        return;
    }
    
    protected boolean runAndReset() {
        if (this.state != 0 || !FutureTask.UNSAFE.compareAndSwapObject(this, FutureTask.runnerOffset, null, Thread.currentThread())) {
            return false;
        }
        boolean n = false;
        int state = this.state;
        int state3;
        try {
            Callable<Object> callable = this.callable;
            if (callable != null && state == 0) {
                try {
                    callable.call();
                    n = true;
                }
                catch (Throwable exception) {
                    this.setException(exception);
                }
            }
            this.runner = null;
            int state2 = this.state;
            if (state2 >= 5) {
                this.handlePossibleCancellationInterrupt(state2);
            }
        }
        finally {
            this.runner = null;
            state3 = this.state;
            if (state3 >= 5) {
                this.handlePossibleCancellationInterrupt(state3);
            }
        }
        return n && state3 == 0;
    }
    
    private void handlePossibleCancellationInterrupt(int n) {
        if (n == 5) {
            while (this.state == 5) {
                Thread.yield();
            }
        }
    }
    
    private void finishCompletion() {
        FutureTask.WaitNode waiters;
        while ((waiters = this.waiters) != null) {
            if (!FutureTask.UNSAFE.compareAndSwapObject(this, FutureTask.waitersOffset, waiters, null)) {
                continue;
            }
            while (true) {
                Thread thread = waiters.thread;
                if (thread != null) {
                    waiters.thread = null;
                    LockSupport.unpark(thread);
                }
                FutureTask.WaitNode next = waiters.next;
                if (next == null) {
                    break;
                }
                waiters.next = null;
                waiters = next;
            }
            break;
        }
        this.done();
        this.callable = null;
    }
    
    private int awaitDone(boolean b, long nanos) throws InterruptedException {
        long n = b ? (System.nanoTime() + nanos) : 0L;
        Object x = null;
        boolean compareAndSwapObject = false;
        while (!Thread.interrupted()) {
            int state = this.state;
            if (state > 1) {
                if (x != null) {
                    x.thread = null;
                }
                return state;
            }
            else if (state == 1) {
                Thread.yield();
            }
            else if (x == null) {
                x = new FutureTask.WaitNode();
            }
            else if (compareAndSwapObject == 0) {
                Unsafe unsafe = FutureTask.UNSAFE;
                long waitersOffset = FutureTask.waitersOffset;
                FutureTask.WaitNode waiters = this.waiters;
                x.next = waiters;
                compareAndSwapObject = unsafe.compareAndSwapObject(this, waitersOffset, waiters, x);
            }
            else if (b) {
                nanos = n - System.nanoTime();
                if (nanos <= 0L) {
                    this.removeWaiter(x);
                    return this.state;
                }
                LockSupport.parkNanos(this, nanos);
            }
            else {
                LockSupport.park(this);
            }
        }
        this.removeWaiter(x);
        throw new InterruptedException();
    }
    
    private void removeWaiter(FutureTask.WaitNode waitNode) {
        if (waitNode != null) {
            waitNode.thread = null;
        Label_0083:
            while (true) {
                FutureTask.WaitNode waitNode2 = null;
                FutureTask.WaitNode next;
                for (FutureTask.WaitNode waiters = this.waiters; waiters != null; waiters = next) {
                    next = waiters.next;
                    if (waiters.thread != null) {
                        waitNode2 = waiters;
                    }
                    else if (waitNode2 != null) {
                        waitNode2.next = next;
                        if (waitNode2.thread == null) {
                            continue Label_0083;
                        }
                    }
                    else if (!FutureTask.UNSAFE.compareAndSwapObject(this, FutureTask.waitersOffset, waiters, next)) {
                        continue Label_0083;
                    }
                }
                break Label_0083;
            }
        }
        else {
            return;
        }
    }
    
    static {
        try {
            UNSAFE = Unsafe.getUnsafe();
            Class<Object> clazz = FutureTask.class;
            stateOffset = FutureTask.UNSAFE.objectFieldOffset(clazz.getDeclaredField("state"));
            runnerOffset = FutureTask.UNSAFE.objectFieldOffset(clazz.getDeclaredField("runner"));
            waitersOffset = FutureTask.UNSAFE.objectFieldOffset(clazz.getDeclaredField("waiters"));
        }
        catch (Exception cause) {
            throw new Error(cause);
        }
    }
    
    static final class WaitNode
    {
        volatile Thread thread;
        volatile FutureTask.WaitNode next;
        
        WaitNode() {
            this.thread = Thread.currentThread();
        }
    }
}
